@@ -30,6 +30,7 @@
 
 #include <unistd.h>
 #include <fcntl.h>
+#include <signal.h>
 #if HAVE_TERMIOS_H
 #include <termios.h>
 #endif
@@ -148,7 +149,7 @@ static int parse_options( int argc, char *argv[] )
 	    error=RETURN_NOERROR;
 	    break;
 	case 'V':
-	    printf("%s (C) 2006-2008 Lingnu Open Source Consulting Ltd.\n"
+	    printf("%s (C) 2006-2011 Lingnu Open Source Consulting Ltd.\n"
 		    "This program is free software, and can be distributed under the terms of the GPL\n"
 		    "See the COPYING file for more information.\n", PACKAGE_STRING );
 	    exit(0);
@@ -173,6 +174,12 @@ int main( int argc, char *argv[] )
         return -(opt_offset+1); // -1 becomes 0, -2 becomes 1 etc.
     }
 
+    if( argc-opt_offset<1 ) {
+	show_help();
+
+        return 0;
+    }
+
     return runprogram( argc-opt_offset, argv+opt_offset );
 }
 
@@ -183,10 +190,15 @@ static int ourtty; // Our own tty
 static int masterpt;
 
 void window_resize_handler(int signum);
+void sigchld_handler(int signum);
 
 int runprogram( int argc, char *argv[] )
 {
     struct winsize ttysize; // The size of our tty
+
+    // We need to interrupt a select with a SIGCHLD. In order to do so, we need a SIGCHLD handler
+    signal( SIGCHLD,sigchld_handler );
+
     // Create a pseudo terminal for our process
     masterpt=posix_openpt(O_RDWR);
 
@@ -196,6 +208,8 @@ int runprogram( int argc, char *argv[] )
 	return RETURN_RUNTIME_ERROR;
     }
 
+    fcntl(masterpt, F_SETFL, O_NONBLOCK);
+
     if( grantpt( masterpt )!=0 ) {
 	perror("Failed to change pseudo terminal's permission");
 
@@ -214,20 +228,45 @@ int runprogram( int argc, char *argv[] )
         ioctl( masterpt, TIOCSWINSZ, &ttysize );
     }
 
+    const char *name=ptsname(masterpt);
+    int slavept;
+    /*
+       Comment no. 3.14159
+
+       This comment documents the history of code.
+
+       We need to open the slavept inside the child process, after "setsid", so that it becomes the controlling
+       TTY for the process. We do not, otherwise, need the file descriptor open. The original approach was to
+       close the fd immediately after, as it is no longer needed.
+
+       It turns out that (at least) the Linux kernel considers a master ptty fd that has no open slave fds
+       to be unused, and causes "select" to return with "error on fd". The subsequent read would fail, causing us
+       to go into an infinite loop. This is a bug in the kernel, as the fact that a master ptty fd has no slaves
+       is not a permenant problem. As long as processes exist that have the slave end as their controlling TTYs,
+       new slave fds can be created by opening /dev/tty, which is exactly what ssh is, in fact, doing.
+
+       Our attempt at solving this problem, then, was to have the child process not close its end of the slave
+       ptty fd. We do, essentially, leak this fd, but this was a small price to pay. This worked great up until
+       openssh version 5.6.
+
+       Openssh version 5.6 looks at all of its open file descriptors, and closes any that it does not know what
+       they are for. While entirely within its prerogative, this breaks our fix, causing sshpass to either
+       hang, or do the infinite loop again.
+
+       Our solution is to keep the slave end open in both parent AND child, at least until the handshake is
+       complete, at which point we no longer need to monitor the TTY anyways.
+     */
+
     int childpid=fork();
     if( childpid==0 ) {
 	// Child
 
 	// Detach us from the current TTY
 	setsid();
+        // This line makes the ptty our controlling tty. We do not otherwise need it open
+        slavept=open(name, O_RDWR );
+        close( slavept );
 	
-	const char *name=ptsname(masterpt);
-	int slavept;
-        slavept=open(name, O_RDWR ); // This line makes the ptty our controlling tty. We do not otherwise need it open
-        // On separate line to confuse gcc into not giving a (technically correct) "unused variable" warning.
-        // This place used to have: close( slavept );
-        // We do not run the above line. See comment 3.1416 later for explanation
-
 	close( masterpt );
 
 	char **new_argv=malloc(sizeof(char *)*(argc+1));
@@ -252,9 +291,22 @@ int runprogram( int argc, char *argv[] )
     }
 	
     // We are the parent
+    slavept=open(name, O_RDWR|O_NOCTTY );
+
     int status=0;
     int terminate=0;
     pid_t wait_id;
+    sigset_t sigmask, sigmask_select;
+
+    // Set the signal mask during the select
+    sigemptyset(&sigmask_select);
+
+    // And during the regular run
+    sigemptyset(&sigmask);
+    sigaddset(&sigmask, SIGCHLD);
+
+    sigprocmask( SIG_SETMASK, &sigmask, NULL );
+
     do {
 	if( !terminate ) {
 	    fd_set readfd;
@@ -262,7 +314,7 @@ int runprogram( int argc, char *argv[] )
 	    FD_ZERO(&readfd);
 	    FD_SET(masterpt, &readfd);
 
-	    int selret=select( masterpt+1, &readfd, NULL, NULL, NULL );
+	    int selret=pselect( masterpt+1, &readfd, NULL, NULL, NULL, &sigmask_select );
 
 	    if( selret>0 ) {
 		if( FD_ISSET( masterpt, &readfd ) ) {
@@ -272,10 +324,16 @@ int runprogram( int argc, char *argv[] )
 
                         // handleoutput returns positive error number in case of some error, and a negative value
                         // if all that happened is that the slave end of the pt is closed.
-                        if( ret>0 )
+                        if( ret>0 ) {
                             close( masterpt ); // Signal ssh that it's controlling TTY is now closed
+                            close(slavept);
+                        }
 
 			terminate=ret;
+
+                        if( terminate ) {
+                            close( slavept );
+                        }
 		    }
 		}
 	    }
@@ -311,16 +369,6 @@ int handleoutput( int fd )
 
     int numread=read(fd, buffer, sizeof(buffer) );
 
-    if( numread<0 ) {
-        // Comment no. 3.1416
-        // Select is doing a horrid job of waking us up at the right time - it wakes up with "read ready" when the slave
-        // end of the pty is closed. This result in an IO error when we perform a read. In the general case, this does
-        // not mean that the master is no more of use, as it may still be that the client will open /dev/tty and send data.
-        // In our case, we keep the slave end open (leaking a file descriptor - the price you pay for API insanity), and so
-        // a failure here suggest ssh is ready to exit.
-        return -1;
-    }
-
     state1=match( compare1, buffer, numread, state1 );
 
     // Are we at a password prompt?
@@ -419,3 +467,8 @@ void window_resize_handler(int signum)
     if( ioctl( ourtty, TIOCGWINSZ, &ttysize )==0 )
         ioctl( masterpt, TIOCSWINSZ, &ttysize );
 }
+
+// Do nothing handler - makes sure the select will terminate if the signal arrives, though.
+void sigchld_handler(int signum)
+{
+}
